/* 
 * Copyright (C) 2003-2004 the Gstreamer project
 * 	Julien Moutte <julien@moutte.net>
 *      Ronald Bultje <rbultje@ronald.bitfreak.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *_
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: bacon-video-widget-gst.c,v 1.87 2004/10/06 12:06:25 rbultje Exp $
 *
 */

#include <config.h>

/* gstgconf */
#include <gst/gconf/gconf.h>

/* GStreamer Interfaces */
#include <gst/xoverlay/xoverlay.h>
#include <gst/navigation/navigation.h>
#include <gst/mixer/mixer.h>
#include <gst/colorbalance/colorbalance.h>

/* system */
#include <string.h>
#include <stdio.h>

/* gtk+/gnome */
#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "bacon-video-widget.h"
#include "baconvideowidget-marshal.h"
#include "scrsaver.h"
#include "video-utils.h"
#include "gststreaminfo.h"

#define DEFAULT_HEIGHT 420
#define DEFAULT_WIDTH 315
#define CONFIG_FILE ".gnome2"G_DIR_SEPARATOR_S"totem_config"
#define DEFAULT_TITLE _("Totem Video Window")

/* Signals */
enum
{
  SIGNAL_ERROR,
  SIGNAL_EOS,
  SIGNAL_TITLE_CHANGE,
  SIGNAL_CHANNELS_CHANGE,
  SIGNAL_TICK,
  SIGNAL_GOT_METADATA,
  SIGNAL_BUFFERING,
  SIGNAL_SPEED_WARNING,
  LAST_SIGNAL
};

/* Enum for none-signal stuff that needs to go through the AsyncQueue */
enum
{
  RATIO = LAST_SIGNAL
};

/* Arguments */
enum
{
  PROP_0,
  PROP_LOGO_MODE,
  PROP_POSITION,
  PROP_CURRENT_TIME,
  PROP_STREAM_LENGTH,
  PROP_PLAYING,
  PROP_SEEKABLE,
  PROP_SHOWCURSOR,
  PROP_MEDIADEV,
  PROP_SHOW_VISUALS
};

struct BaconVideoWidgetPrivate
{
  double display_ratio;

  GstElement *play;
  guint update_id;
  GstMixer *mixer;
  GstMixerTrack *mixer_track;
  GstXOverlay *xoverlay;
  GstColorBalance *balance;

  GdkPixbuf *logo_pixbuf;

  gboolean media_has_video, media_has_audio;

  gint64 stream_length;
  gint64 current_time_nanos;
  gint64 current_time;
  float current_position;

  GstTagList *tagcache;

  char *last_error_message;

  GdkWindow *video_window;
  GtkAllocation video_window_allocation;

  /* Visual effects */
  GList *vis_plugins_list;
  gboolean show_vfx;
  gboolean using_vfx;
  GstElement *vis_element;

  /* Other stuff */
  int xpos, ypos;
  gboolean logo_mode;
  gboolean cursor_shown;
  gboolean auto_resize;
  
  GAsyncQueue *queue;
  
  gint video_width;
  gint video_height;
  gdouble video_fps;

  guint init_width;
  guint init_height;
  
  char *mrl;
  char *media_device;
};

enum {
  ASYNC_VIDEO_SIZE,
  ASYNC_ERROR,
  ASYNC_FOUND_TAG,
  ASYNC_EOS
};

typedef struct _BVWSignal BVWSignal;

struct _BVWSignal
{
  gint signal_id;
  union
  {
    struct
    {
      gint width;
      gint height;
    } video_size;
    struct
    {
      GstElement *element;
      GError *error;
      char *debug_message;
    } error;
    struct
    {
      GstElement *source;
      GstTagList *tag_list;
    } found_tag;
  } signal_data;
};

static void bacon_video_widget_set_property (GObject * object,
					     guint property_id,
					     const GValue * value,
					     GParamSpec * pspec);
static void bacon_video_widget_get_property (GObject * object,
					     guint property_id,
					     GValue * value,
					     GParamSpec * pspec);

static void bacon_video_widget_finalize (GObject * object);

static GtkWidgetClass *parent_class = NULL;

static int bvw_table_signals[LAST_SIGNAL] = { 0 };

static void
bacon_video_widget_realize (GtkWidget * widget)
{
  BaconVideoWidget *bvw = BACON_VIDEO_WIDGET (widget);
  GdkWindowAttr attributes;
  gint attributes_mask;

  /* Creating our widget's window */
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= GDK_EXPOSURE_MASK;
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
				   &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, widget);
  gdk_window_show (widget->window);

  /* Creating our video window */
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = 0;
  attributes.y = 0;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.event_mask = GDK_EXPOSURE_MASK;
  attributes_mask = GDK_WA_X | GDK_WA_Y;

  bvw->priv->video_window = gdk_window_new (widget->window,
				            &attributes, attributes_mask);
  gdk_window_set_user_data (bvw->priv->video_window, widget);
  gdk_window_show (bvw->priv->video_window);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
}

static gboolean
bacon_video_widget_expose_event (GtkWidget *widget, GdkEventExpose *event)
{
  BaconVideoWidget *bvw = BACON_VIDEO_WIDGET (widget);
  XID window;

  if (event && event->count > 0)
    return TRUE;

  g_return_val_if_fail (bvw->priv->xoverlay != NULL &&
			GST_IS_X_OVERLAY (bvw->priv->xoverlay),
			FALSE);

  if (GST_STATE (bvw->priv->play) >= GST_STATE_PAUSED) {
    window = GDK_WINDOW_XWINDOW (bvw->priv->video_window);
  } else {
    window = 0;
  }

  gdk_draw_rectangle (widget->window, widget->style->black_gc, TRUE, 0, 0,
		      widget->allocation.width, widget->allocation.height);
  gdk_draw_rectangle (bvw->priv->video_window, widget->style->black_gc, TRUE,
		      bvw->priv->video_window_allocation.x,
		      bvw->priv->video_window_allocation.y,
		      bvw->priv->video_window_allocation.width,
		      bvw->priv->video_window_allocation.height);
  gst_x_overlay_set_xwindow_id (bvw->priv->xoverlay, window);

  if (GST_STATE (bvw->priv->play) >= GST_STATE_PAUSED) {
    gst_x_overlay_expose (bvw->priv->xoverlay);
  } else if (bvw->priv->logo_pixbuf != NULL) {
    /* draw logo here */
    GdkPixbuf *logo;
    gfloat width = gdk_pixbuf_get_width (bvw->priv->logo_pixbuf),
           height = gdk_pixbuf_get_height (bvw->priv->logo_pixbuf);
    gfloat ratio;

    if ((gfloat) widget->allocation.width /
        gdk_pixbuf_get_width (bvw->priv->logo_pixbuf) >
        (gfloat) widget->allocation.height /
        gdk_pixbuf_get_height (bvw->priv->logo_pixbuf)) {
      ratio = (gfloat) widget->allocation.height /
	      gdk_pixbuf_get_height (bvw->priv->logo_pixbuf);
    } else {
      ratio = (gfloat) widget->allocation.width /
	      gdk_pixbuf_get_width (bvw->priv->logo_pixbuf);
    }
    width *= ratio;
    height *= ratio;

    logo = gdk_pixbuf_scale_simple (bvw->priv->logo_pixbuf,
				    width, height,
				    GDK_INTERP_BILINEAR);

    gdk_draw_pixbuf (GDK_DRAWABLE (bvw->priv->video_window),
		     widget->style->fg_gc[0], logo,
		     0, 0, 0, 0, width, height,
		     GDK_RGB_DITHER_NONE, 0, 0);

    gdk_pixbuf_unref (logo);
  }
  
  return TRUE;
}

static void
bacon_video_widget_size_request (GtkWidget * widget,
				 GtkRequisition * requisition)
{
  BaconVideoWidget *bvw = BACON_VIDEO_WIDGET (widget);

  /* FIXME: intial width, this is minimal, but not "recommended"
   * (media size) */
  if (GST_STATE (bvw->priv->play) >= GST_STATE_PAUSED) {
    requisition->width = bvw->priv->video_width;
    requisition->height = bvw->priv->video_height;
  } else {
    if (bvw->priv->logo_pixbuf != NULL) {
      requisition->width = gdk_pixbuf_get_width (bvw->priv->logo_pixbuf);
      requisition->height = gdk_pixbuf_get_height (bvw->priv->logo_pixbuf);
    } else {
      requisition->width = bvw->priv->init_width;
      requisition->height = bvw->priv->init_height;
    }
  }
}

static void
bacon_video_widget_size_allocate (GtkWidget * widget,
				  GtkAllocation * allocation)
{
  BaconVideoWidget *bvw = BACON_VIDEO_WIDGET (widget);

  g_return_if_fail (widget != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (widget));

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget)) {
    gfloat width, height, ratio;

    gdk_window_move_resize (widget->window,
                            allocation->x, allocation->y,
                            allocation->width, allocation->height);

    /* resize video_window */
    if (GST_STATE (bvw->priv->play) >= GST_STATE_PAUSED) {
      width = bvw->priv->video_width;
      height = bvw->priv->video_height;
    } else if (bvw->priv->logo_pixbuf) {
      width = gdk_pixbuf_get_width (bvw->priv->logo_pixbuf);
      height = gdk_pixbuf_get_height (bvw->priv->logo_pixbuf);
    } else {
      width = bvw->priv->init_width;
      height = bvw->priv->init_height;
    }

    if ((gfloat) allocation->width / width >
        (gfloat) allocation->height / height) {
      ratio = (gfloat) allocation->height / height;
    } else {
      ratio = (gfloat) allocation->width / width;
    }

    width *= ratio;
    height *= ratio;

    bvw->priv->video_window_allocation.width = width;
    bvw->priv->video_window_allocation.height = height;
    bvw->priv->video_window_allocation.x = (allocation->width - width) / 2;
    bvw->priv->video_window_allocation.y = (allocation->height - height) / 2;
    gdk_window_move_resize (bvw->priv->video_window,
                            (allocation->width - width) / 2,
			    (allocation->height - height) / 2,
                            width, height);
  }
}

static void
bacon_video_widget_class_init (BaconVideoWidgetClass * klass)
{
  GObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GObjectClass *) klass;
  widget_class = (GtkWidgetClass *) klass;

  parent_class = gtk_type_class (gtk_box_get_type ());

  /* GtkWidget */
  widget_class->size_request = bacon_video_widget_size_request;
  widget_class->size_allocate = bacon_video_widget_size_allocate;
  widget_class->realize = bacon_video_widget_realize;
  widget_class->expose_event = bacon_video_widget_expose_event;

  /* FIXME:
   * - once I've re-added DVD supports, I want to add event handlers
   *    ( keys, mouse buttons, mouse motion) here, too.
   */
  
  /* GObject */
  object_class->set_property = bacon_video_widget_set_property;
  object_class->get_property = bacon_video_widget_get_property;
  object_class->finalize = bacon_video_widget_finalize;

  /* Properties */
  g_object_class_install_property (object_class, PROP_LOGO_MODE,
				   g_param_spec_boolean ("logo_mode", NULL,
							 NULL, FALSE,
							 G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_POSITION,
				   g_param_spec_int ("position", NULL, NULL,
						     0, G_MAXINT, 0,
						     G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_STREAM_LENGTH,
				   g_param_spec_int64 ("stream_length", NULL,
						     NULL, 0, G_MAXINT64, 0,
						     G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_PLAYING,
				   g_param_spec_boolean ("playing", NULL,
							 NULL, FALSE,
							 G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_SEEKABLE,
				   g_param_spec_boolean ("seekable", NULL,
							 NULL, FALSE,
							 G_PARAM_READABLE));
  g_object_class_install_property (object_class, PROP_SHOWCURSOR,
				   g_param_spec_boolean ("showcursor", NULL,
							 NULL, FALSE,
							 G_PARAM_READWRITE));
  g_object_class_install_property (object_class, PROP_MEDIADEV,
				   g_param_spec_string ("mediadev", NULL,
							NULL, FALSE,
							G_PARAM_WRITABLE));
  g_object_class_install_property (object_class, PROP_SHOW_VISUALS,
				   g_param_spec_boolean ("showvisuals", NULL,
							 NULL, FALSE,
							 G_PARAM_WRITABLE));

  /* Signals */
  bvw_table_signals[SIGNAL_ERROR] =
    g_signal_new ("error",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, error),
		  NULL, NULL,
		  baconvideowidget_marshal_VOID__STRING_BOOLEAN_BOOLEAN,
		  G_TYPE_NONE, 3, G_TYPE_STRING,
		  G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);

  bvw_table_signals[SIGNAL_EOS] =
    g_signal_new ("eos",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, eos),
		  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

  bvw_table_signals[SIGNAL_GOT_METADATA] =
    g_signal_new ("got-metadata",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, got_metadata),
		  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

  bvw_table_signals[SIGNAL_TITLE_CHANGE] =
    g_signal_new ("title-change",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, title_change),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__STRING,
		  G_TYPE_NONE, 1, G_TYPE_STRING);

  bvw_table_signals[SIGNAL_CHANNELS_CHANGE] =
    g_signal_new ("channels-change",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, channels_change),
		  NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

  bvw_table_signals[SIGNAL_TICK] =
    g_signal_new ("tick",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, tick),
		  NULL, NULL,
		  baconvideowidget_marshal_VOID__INT64_INT64_FLOAT_BOOLEAN,
		  G_TYPE_NONE, 4, G_TYPE_INT64, G_TYPE_INT64, G_TYPE_FLOAT,
                  G_TYPE_BOOLEAN);

  bvw_table_signals[SIGNAL_BUFFERING] =
    g_signal_new ("buffering",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, buffering),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);

  bvw_table_signals[SIGNAL_SPEED_WARNING] =
    g_signal_new ("speed-warning",
		  G_TYPE_FROM_CLASS (object_class),
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (BaconVideoWidgetClass, speed_warning),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);
}

static void
bacon_video_widget_init (BaconVideoWidget * bvw)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  GTK_WIDGET_SET_FLAGS (GTK_WIDGET (bvw), GTK_CAN_FOCUS);
  GTK_WIDGET_SET_FLAGS (GTK_WIDGET (bvw), GTK_NO_WINDOW);
  GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (bvw), GTK_DOUBLE_BUFFERED);

  bvw->priv = g_new0 (BaconVideoWidgetPrivate, 1);
  
  bvw->priv->queue = g_async_queue_new ();
  bvw->priv->update_id = 0;
  bvw->priv->tagcache = NULL;
}

static void
shrink_toplevel (BaconVideoWidget * bvw)
{
  GtkWidget *toplevel;
  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (bvw));
  gtk_window_resize (GTK_WINDOW (toplevel), 1, 1);
  gtk_widget_queue_resize (GTK_WIDGET (bvw));
}

static gboolean
bacon_video_widget_signal_idler (BaconVideoWidget *bvw)
{
  BVWSignal *signal = NULL;
  gint queue_length;
  
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  
  signal = g_async_queue_try_pop (bvw->priv->queue);
  if (!signal)
    return FALSE;
  
  switch (signal->signal_id)
    {
      case ASYNC_VIDEO_SIZE:
        {
          GList *streaminfo = NULL;

          g_object_get (G_OBJECT (bvw->priv->play), "stream-info",
			&streaminfo, NULL);
          for ( ; streaminfo != NULL; streaminfo = streaminfo->next) {
            GstStreamInfo *info = streaminfo->data;

            switch (info->type) {
              case GST_STREAM_TYPE_VIDEO: {
                GstStructure *s;

                s = gst_caps_get_structure (GST_PAD_CAPS (info->pad), 0);
                if (s) {
                  gst_structure_get_double (s, "framerate", &bvw->priv->video_fps);
                  gst_structure_get_int (s, "width", &bvw->priv->video_width);
                  gst_structure_get_int (s, "height", &bvw->priv->video_height);
                  /* FIXME: a-s-r */
                }
                break;
              }
              default:
                break;
            }
          }

          g_signal_emit (G_OBJECT (bvw), bvw_table_signals[SIGNAL_GOT_METADATA], 0, NULL);

          if (bvw->priv->auto_resize) {
            shrink_toplevel (bvw);
          } else {
            bacon_video_widget_size_allocate (GTK_WIDGET (bvw),
					      &GTK_WIDGET (bvw)->allocation);
          }

          break;
        }
      case ASYNC_ERROR:
        {
          char *error_message = NULL;
          gboolean emit = TRUE;
          
          if (signal->signal_data.error.error)
            error_message = signal->signal_data.error.error->message;

          if (bvw->priv->last_error_message) {
            /* Let's check the latest error message */
            if (g_ascii_strcasecmp (error_message,
                                    bvw->priv->last_error_message) == 0)
              emit = FALSE;
          }

          if (emit) {
            g_signal_emit (G_OBJECT (bvw),
                           bvw_table_signals[SIGNAL_ERROR], 0,
			   error_message, TRUE, FALSE);
            
            /* Keep a copy of the last emitted message */
            if (bvw->priv->last_error_message)
              g_free (bvw->priv->last_error_message);
            bvw->priv->last_error_message = g_strdup (error_message);
          }
          
          /* Cleaning the error infos */
          if (signal->signal_data.error.error)
            g_error_free (signal->signal_data.error.error);
          if (signal->signal_data.error.debug_message)
            g_free (signal->signal_data.error.debug_message);
          
          break;
        }
      case ASYNC_FOUND_TAG:
        {
          GstTagList *tag_list = signal->signal_data.found_tag.tag_list;

          if (tag_list) {
            if (bvw->priv->tagcache) {
              GstTagList *taglist;

              taglist = gst_tag_list_merge (bvw->priv->tagcache, tag_list,
					    GST_TAG_MERGE_APPEND);
              gst_tag_list_free (tag_list);
              gst_tag_list_free (bvw->priv->tagcache);
              bvw->priv->tagcache = taglist;
            } else {
              bvw->priv->tagcache = tag_list;
            }
            g_signal_emit (G_OBJECT (bvw), bvw_table_signals[SIGNAL_GOT_METADATA],
                           0, NULL);
          }
          break;
        }
      case ASYNC_EOS:
        {
          gst_element_set_state (GST_ELEMENT (bvw->priv->play),
                                 GST_STATE_READY);
          g_signal_emit (G_OBJECT (bvw), bvw_table_signals[SIGNAL_EOS], 0, NULL);
          break;
        }
      default:
        break;
    }
    
  g_free (signal);
  queue_length = g_async_queue_length (bvw->priv->queue);

  return (queue_length > 0);
}

static void
got_found_tag (GstElement *play, GstElement *source,
               GstTagList *tag_list, BaconVideoWidget * bvw)
{
  BVWSignal *signal;
  
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  
  signal = g_new0 (BVWSignal, 1);
  signal->signal_id = ASYNC_FOUND_TAG;
  signal->signal_data.found_tag.source = source;
  signal->signal_data.found_tag.tag_list = gst_tag_list_copy (tag_list);

  g_async_queue_push (bvw->priv->queue, signal);

  g_idle_add ((GSourceFunc) bacon_video_widget_signal_idler, bvw);
}

static void
got_video_size (GstElement * play, guint width, guint height,
		BaconVideoWidget * bvw)
{
  BVWSignal *signal;
  
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  
  signal = g_new0 (BVWSignal, 1);
  signal->signal_id = ASYNC_VIDEO_SIZE;
  signal->signal_data.video_size.width = width;
  signal->signal_data.video_size.height = height;

  g_async_queue_push (bvw->priv->queue, signal);

  g_idle_add ((GSourceFunc) bacon_video_widget_signal_idler, bvw);
}

static void
got_eos (GstElement * play, BaconVideoWidget * bvw)
{
  BVWSignal *signal;
  
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  
  signal = g_new0 (BVWSignal, 1);
  signal->signal_id = ASYNC_EOS;

  g_async_queue_push (bvw->priv->queue, signal);

  g_idle_add ((GSourceFunc) bacon_video_widget_signal_idler, bvw);
}

static void
got_stream_length (GstElement * play, gint64 length_nanos,
                   BaconVideoWidget * bvw)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  bvw->priv->stream_length = (gint64) length_nanos / GST_MSECOND;
  
  g_signal_emit (G_OBJECT (bvw), bvw_table_signals[SIGNAL_GOT_METADATA], 0, NULL);
}

static void
got_time_tick (GstElement * play, gint64 time_nanos, BaconVideoWidget * bvw)
{
  gboolean seekable;

  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  if (bvw->priv->logo_mode != FALSE)
    return;

  bvw->priv->current_time_nanos = time_nanos;

  bvw->priv->current_time = (gint64) time_nanos / GST_MSECOND;

  if (bvw->priv->stream_length == 0)
    bvw->priv->current_position = 0;
  else
    {
      bvw->priv->current_position =
        (float) bvw->priv->current_time / bvw->priv->stream_length;
    }
  
  seekable = bacon_video_widget_is_seekable (bvw);

  g_signal_emit (G_OBJECT (bvw),
                 bvw_table_signals[SIGNAL_TICK], 0,
                 bvw->priv->current_time, bvw->priv->stream_length,
                 bvw->priv->current_position,
                 seekable);
}

static gboolean
cb_iterate (BaconVideoWidget *bvw)
{
  GstFormat fmt = GST_FORMAT_TIME;
  gint64 value;
  gboolean res;

  if (!GST_FLAG_IS_SET (bvw->priv->play, GST_BIN_SELF_SCHEDULABLE)) {
    res = gst_bin_iterate (GST_BIN (bvw->priv->play));
  } else {
    g_usleep (100);
    res = (gst_element_get_state (bvw->priv->play) == GST_STATE_PLAYING);
  }

  /* check length/pos of stream */
  if (!bvw->priv->stream_length) {
    if (gst_element_query (GST_ELEMENT (bvw->priv->play),
			   GST_QUERY_TOTAL, &fmt, &value)) {
      got_stream_length (GST_ELEMENT (bvw->priv->play),
			 value, bvw);
    }
  }
  if (gst_element_query (GST_ELEMENT (bvw->priv->play),
			 GST_QUERY_POSITION, &fmt, &value)) {
    got_time_tick (GST_ELEMENT (bvw->priv->play),
		   value, bvw);
  }

  if (!res)
    bvw->priv->update_id = 0;

  return res;
}

static void
state_change (GstElement *play, GstElementState old_state,
	      GstElementState new_state, BaconVideoWidget *bvw)
{
  if (old_state == GST_STATE_PLAYING) {
    if (bvw->priv->update_id != 0) {
      g_source_remove (bvw->priv->update_id);
      bvw->priv->update_id = 0;
    }
  } else if (new_state == GST_STATE_PLAYING) {
    if (bvw->priv->update_id != 0)
      g_source_remove (bvw->priv->update_id);
    bvw->priv->update_id = g_idle_add ((GSourceFunc) cb_iterate, bvw);
  }

  if (old_state <= GST_STATE_READY &&
      new_state >= GST_STATE_PAUSED) {
    GList *streaminfo = NULL;

    g_object_get (G_OBJECT (bvw->priv->play), "stream-info",
		  &streaminfo, NULL);
    for ( ; streaminfo != NULL; streaminfo = streaminfo->next) {
      GstStreamInfo *info = streaminfo->data;

      switch (info->type) {
        case GST_STREAM_TYPE_AUDIO:
          bvw->priv->media_has_audio = TRUE;
          break;
        case GST_STREAM_TYPE_VIDEO:
          bvw->priv->media_has_video = TRUE;
          break;
        default:
          break;
      }
    }
  } else if (new_state <= GST_STATE_READY &&
             old_state >= GST_STATE_PAUSED) {
    bvw->priv->media_has_video = FALSE;
    bvw->priv->media_has_audio = FALSE;

    /* clean metadata cache */
    if (bvw->priv->tagcache)
      {
        gst_tag_list_free (bvw->priv->tagcache);
        bvw->priv->tagcache = NULL;
      }
  }
}

static void
got_error (GstElement *play, GstElement *orig, GError *error,
           gchar *debug, BaconVideoWidget * bvw)
{
  BVWSignal *signal;
  
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  /* the handler will catch it */
  if (GST_STATE (play) != GST_STATE_PLAYING)
    return;
  
  signal = g_new0 (BVWSignal, 1);
  signal->signal_id = ASYNC_ERROR;
  signal->signal_data.error.element = orig;
  signal->signal_data.error.error = g_error_copy (error);
  if (debug)
    signal->signal_data.error.debug_message = g_strdup (debug);

  g_async_queue_push (bvw->priv->queue, signal);

  g_idle_add ((GSourceFunc) bacon_video_widget_signal_idler, bvw);
}

static void
bacon_video_widget_finalize (GObject * object)
{
  BaconVideoWidget *bvw = (BaconVideoWidget *) object;

  if (bvw->priv->media_device)
    {
      g_free (bvw->priv->media_device);
      bvw->priv->media_device = NULL;
    }
    
  if (bvw->priv->mrl)
    {
      g_free (bvw->priv->mrl);
      bvw->priv->mrl = NULL;
    }
    
  if (bvw->priv->queue)
    {
      g_async_queue_unref (bvw->priv->queue);
      bvw->priv->queue = NULL;
    }

  if (bvw->priv->vis_plugins_list)
    {
      g_list_foreach (bvw->priv->vis_plugins_list, (GFunc) g_free, NULL);
      g_list_free (bvw->priv->vis_plugins_list);
      bvw->priv->vis_plugins_list = NULL;
    }

  if (bvw->priv->play != NULL && GST_IS_ELEMENT (bvw->priv->play))
    {
      gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_READY);
      gst_object_unref (GST_OBJECT (bvw->priv->play));
      bvw->priv->play = NULL;
    }

  if (bvw->priv->update_id)
    {
      g_source_remove (bvw->priv->update_id);
      bvw->priv->update_id = 0;
    }

  if (bvw->priv->tagcache)
    {
      gst_tag_list_free (bvw->priv->tagcache);
      bvw->priv->tagcache = NULL;
    }
}

static void
bacon_video_widget_set_property (GObject * object, guint property_id,
                                 const GValue * value, GParamSpec * pspec)
{
}

static void
bacon_video_widget_get_property (GObject * object, guint property_id,
                                 GValue * value, GParamSpec * pspec)
{
}

/* ============================================================= */
/*                                                               */
/*                       Public Methods                          */
/*                                                               */
/* ============================================================= */

char *
bacon_video_widget_get_backend_name (BaconVideoWidget * bvw)
{
  guint major, minor, micro;

  gst_version (&major, &minor, &micro);

  return g_strdup_printf ("GStreamer version %d.%d.%d", major, minor, micro);
}

int
bacon_video_widget_get_subtitle (BaconVideoWidget * bvw)
{
  return -1;
}

void
bacon_video_widget_set_subtitle (BaconVideoWidget * bvw, int subtitle)
{
}

GList * bacon_video_widget_get_subtitles (BaconVideoWidget * bvw)
{
  return NULL;
}

GList * bacon_video_widget_get_languages (BaconVideoWidget * bvw)
{
  return NULL;
}

int
bacon_video_widget_get_language (BaconVideoWidget * bvw)
{
  return -1;
}

void
bacon_video_widget_set_language (BaconVideoWidget * bvw, int language)
{
}

int
bacon_video_widget_get_connection_speed (BaconVideoWidget * bvw)
{
  return 0;
}

void
bacon_video_widget_set_connection_speed (BaconVideoWidget * bvw, int speed)
{
}

void
bacon_video_widget_set_deinterlacing (BaconVideoWidget * bvw,
				      gboolean deinterlace)
{
}

gboolean
bacon_video_widget_get_deinterlacing (BaconVideoWidget * bvw)
{
  return FALSE;
}

gboolean
bacon_video_widget_set_tv_out (BaconVideoWidget * bvw, TvOutType tvout)
{
  return FALSE;
}

TvOutType
bacon_video_widget_get_tv_out (BaconVideoWidget * bvw)
{
  return TV_OUT_NONE;
}

BaconVideoWidgetAudioOutType
bacon_video_widget_get_audio_out_type (BaconVideoWidget *bvw)
{
  return BVW_AUDIO_SOUND_STEREO;
}

void
bacon_video_widget_set_audio_out_type (BaconVideoWidget *bvw,
                                       BaconVideoWidgetAudioOutType type)
{
  
}

/* =========================================== */
/*                                             */
/*               Play/Pause, Stop              */
/*                                             */
/* =========================================== */

gboolean
bacon_video_widget_open (BaconVideoWidget * bvw, const gchar * mrl,
			 GError ** error)
{
  gboolean ret;

  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (mrl != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (bvw->priv->play != NULL, FALSE);
  g_return_val_if_fail (bvw->priv->mrl == NULL, FALSE);

  g_free (bvw->priv->mrl);
  bvw->priv->mrl = g_strdup (mrl);

  gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_READY);

  /* Resetting last_error_message to NULL */
  if (bvw->priv->last_error_message)
    {
      g_free (bvw->priv->last_error_message);
      bvw->priv->last_error_message = NULL;
    }


  bvw->priv->media_has_video = FALSE;
  bvw->priv->stream_length = 0;

  g_object_set (G_OBJECT (bvw->priv->play), "uri",
		bvw->priv->mrl, NULL);

  ret = (gst_element_set_state (bvw->priv->play, GST_STATE_PAUSED) ==
      GST_STATE_SUCCESS);
  if (!ret)
    {
      g_set_error (error, 0, 0, bvw->priv->last_error_message ?
          bvw->priv->last_error_message : "Failed to open; reason unknown");
    }

  return ret;
}

gboolean
bacon_video_widget_play (BaconVideoWidget * bvw, GError ** error)
{
  gboolean ret;

  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  /* Resetting last_error_message to NULL */
  if (bvw->priv->last_error_message)
    {
      g_free (bvw->priv->last_error_message);
      bvw->priv->last_error_message = NULL;
    }

  ret = (gst_element_set_state (GST_ELEMENT (bvw->priv->play),
      GST_STATE_PLAYING) == GST_STATE_SUCCESS);
  if (!ret)
    {
      g_set_error (error, 0, 0, bvw->priv->last_error_message ?
          bvw->priv->last_error_message : "Failed to play; reason unknown");
    }

  return ret;
}

gboolean
bacon_video_widget_seek (BaconVideoWidget *bvw, float position, GError **gerror)
{
  gint64 seek_time, length_nanos;

  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  /* Resetting last_error_message to NULL */
  if (bvw->priv->last_error_message)
    {
      g_free (bvw->priv->last_error_message);
      bvw->priv->last_error_message = NULL;
    }

  length_nanos = (gint64) (bvw->priv->stream_length * GST_MSECOND);
  seek_time = (gint64) (length_nanos * position);

  gst_element_seek (bvw->priv->play, GST_SEEK_METHOD_SET |
		    GST_SEEK_FLAG_FLUSH | GST_FORMAT_TIME,
		    seek_time);

  return TRUE;
}

gboolean
bacon_video_widget_seek_time (BaconVideoWidget *bvw, gint64 time, GError **gerror)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  /* Resetting last_error_message to NULL */
  if (bvw->priv->last_error_message)
    {
      g_free (bvw->priv->last_error_message);
      bvw->priv->last_error_message = NULL;
    }

  gst_element_seek (bvw->priv->play, GST_SEEK_METHOD_SET |
		    GST_SEEK_FLAG_FLUSH | GST_FORMAT_TIME,
		    time * GST_MSECOND);

  return TRUE;
}

void
bacon_video_widget_stop (BaconVideoWidget * bvw)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_READY);
}

void
bacon_video_widget_close (BaconVideoWidget * bvw)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_READY);
  
  if (bvw->priv->mrl) {
    g_free (bvw->priv->mrl);
    bvw->priv->mrl = NULL;
  }
}

void
bacon_video_widget_dvd_event (BaconVideoWidget * bvw,
			      BaconVideoWidgetDVDEvent type)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));
}

void
bacon_video_widget_set_logo (BaconVideoWidget * bvw, gchar * filename)
{
  GError *error = NULL;
  
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  bvw->priv->logo_pixbuf = gdk_pixbuf_new_from_file (filename, &error);

  if (error) {
    g_warning ("An error occured trying to open logo %s: %s",
               filename, error->message);
    g_error_free (error);
  } else {
    shrink_toplevel (bvw);
  }
}

void
bacon_video_widget_set_logo_mode (BaconVideoWidget * bvw, gboolean logo_mode)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  /* this probably tells us to show logo or not? We already know
   * that from the media...? */
  bvw->priv->logo_mode = logo_mode;
}

gboolean
bacon_video_widget_get_logo_mode (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);

  return bvw->priv->logo_mode;
}

void
bacon_video_widget_pause (BaconVideoWidget * bvw)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_PAUSED);
}

void
bacon_video_widget_set_proprietary_plugins_path (BaconVideoWidget * bvw,
						 const char *path)
{
}

gboolean
bacon_video_widget_can_set_volume (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);
  
  return TRUE;
}

void
bacon_video_widget_set_volume (BaconVideoWidget * bvw, int volume)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  g_object_set (G_OBJECT (bvw->priv->play), "volume",
		(gdouble) (1. * volume / 100), NULL);
}

int
bacon_video_widget_get_volume (BaconVideoWidget * bvw)
{
  gdouble vol;

  g_return_val_if_fail (bvw != NULL, -1);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), -1);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), -1);

  g_object_get (G_OBJECT (bvw->priv->play), "volume", &vol, NULL);

  return (gint) (vol * 100 + 0.5);
}

gboolean
bacon_video_widget_fullscreen_mode_available (BaconVideoWidget *bvw,
		TvOutType tvout) 
{
	switch(tvout) {
	case TV_OUT_NONE:
		/* Asume that ordinary fullscreen always works */
		return TRUE;
	case TV_OUT_NVTV_NTSC:
	case TV_OUT_NVTV_PAL:
		return FALSE;
	case TV_OUT_DXR3:
		/* FIXME: Add DXR3 detection code */
		return FALSE;
	}
	return FALSE;
}

void
bacon_video_widget_set_fullscreen (BaconVideoWidget * bvw,
				   gboolean fullscreen)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
}

void
bacon_video_widget_set_show_cursor (BaconVideoWidget * bvw,
				    gboolean show_cursor)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  if (show_cursor == FALSE)
  {
    eel_gdk_window_set_invisible_cursor (bvw->priv->video_window);
  } else {
    gdk_window_set_cursor (bvw->priv->video_window, NULL);
  }

  bvw->priv->cursor_shown = show_cursor;
}

gboolean
bacon_video_widget_get_show_cursor (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);

  return bvw->priv->cursor_shown;
}

void
bacon_video_widget_set_media_device (BaconVideoWidget * bvw, const char *path)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));
  
  if (bvw->priv->media_device)
    g_free (bvw->priv->media_device);
  
  bvw->priv->media_device = g_strdup (path);
}

gboolean
bacon_video_widget_set_show_visuals (BaconVideoWidget * bvw,
				     gboolean show_visuals)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  return FALSE;
#if 0
  if (!bvw->priv->media_has_video)
    {
      if (!show_visuals)
	gst_video_widget_set_logo_focus (bvw->priv->vw, TRUE);
      else
	gst_video_widget_set_logo_focus (bvw->priv->vw, FALSE);

      gst_play_connect_visualization (bvw->priv->play, show_visuals);
    }

  return TRUE;
#endif
}

GList *
bacon_video_widget_get_visuals_list (BaconVideoWidget * bvw)
{
  GList *pool_registries = NULL, *registries = NULL, *vis_plugins_list = NULL;
  GList *plugins = NULL, *features = NULL;

  g_return_val_if_fail (bvw != NULL, NULL);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), NULL);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), NULL);

  /* Cache */
  if (bvw->priv->vis_plugins_list)
    {
      return bvw->priv->vis_plugins_list;
    }

  pool_registries = gst_registry_pool_list ();

  registries = pool_registries;

  while (registries)
    {
      GstRegistry *registry = GST_REGISTRY (registries->data);

      plugins = registry->plugins;

      while (plugins)
	{
	  GstPlugin *plugin = GST_PLUGIN (plugins->data);

	  features = gst_plugin_get_feature_list (plugin);

	  while (features)
	    {
	      GstPluginFeature *feature = GST_PLUGIN_FEATURE (features->data);

	      if (GST_IS_ELEMENT_FACTORY (feature))
		{
		  GstElementFactory *factory = GST_ELEMENT_FACTORY (feature);
		  if (g_strrstr (factory->details.klass, "Visualization"))
		    {
		      vis_plugins_list = g_list_append (vis_plugins_list,
							(char *)
							g_strdup
							(GST_OBJECT_NAME
							 (factory)));
		    }
		}
	      features = g_list_next (features);
	    }

	  plugins = g_list_next (plugins);
	}

      registries = g_list_next (registries);
    }

  g_list_free (pool_registries);
  pool_registries = NULL;

  bvw->priv->vis_plugins_list = vis_plugins_list;

  return vis_plugins_list;
}

gboolean
bacon_video_widget_set_visuals (BaconVideoWidget * bvw, const char *name)
{
  //gboolean paused = FALSE;

  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  return FALSE;
#if 0
  if (GST_STATE (GST_ELEMENT (bvw->priv->play)) == GST_STATE_PLAYING)
    {
      gst_element_set_state (GST_ELEMENT (bvw->priv->play), GST_STATE_PAUSED);
      paused = TRUE;
    }

  bvw->priv->vis_element = gst_element_factory_make (name,
                                                     "vis_plugin_element");
  if (GST_IS_ELEMENT (bvw->priv->vis_element))
    {
      gst_play_set_visualization (bvw->priv->play, bvw->priv->vis_element);
      if (paused)
	{
	  gst_play_seek_to_time (bvw->priv->play,
				 bvw->priv->current_time_nanos);
	  gst_element_set_state (GST_ELEMENT (bvw->priv->play),
                                 GST_STATE_PLAYING);
	}

      return FALSE;
    }
  else
    {
      return FALSE;
    }
#endif
}

void
bacon_video_widget_set_visuals_quality (BaconVideoWidget * bvw,
					VisualsQuality quality)
{
  int fps, w, h;
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  if (bvw->priv->vis_element == NULL)
    return;

  switch (quality)
    {
    case VISUAL_SMALL:
      fps = 15;
      w = 320;
      h = 240;
      break;
    case VISUAL_NORMAL:
      fps = 25;
      w = 320;
      h = 240;
      break;
    case VISUAL_LARGE:
      fps = 25;
      w = 640;
      h = 480;
      break;
    case VISUAL_EXTRA_LARGE:
      fps = 30;
      w = 800;
      h = 600;
      break;
    default:
      /* shut up warnings */
      fps = w = h = 0;
      g_assert_not_reached ();
  }

}

gboolean
bacon_video_widget_get_auto_resize (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);

  return bvw->priv->auto_resize;
}

void
bacon_video_widget_set_auto_resize (BaconVideoWidget * bvw,
				    gboolean auto_resize)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));

  bvw->priv->auto_resize = auto_resize;

  /* this will take effect when the next media file loads */
}

void
bacon_video_widget_set_aspect_ratio (BaconVideoWidget *bvw,
				BaconVideoWidgetAspectRatio ratio)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
}

BaconVideoWidgetAspectRatio
bacon_video_widget_get_aspect_ratio (BaconVideoWidget *bvw)
{
  g_return_val_if_fail (bvw != NULL, 0);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), 0);

  return BVW_RATIO_AUTO;
}

void
bacon_video_widget_set_scale_ratio (BaconVideoWidget * bvw, gfloat ratio)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

//  gst_video_widget_set_scale (bvw->priv->vw, ratio);
//  gst_video_widget_set_scale_override (bvw->priv->vw, TRUE);
  shrink_toplevel (bvw);
}

int
bacon_video_widget_get_video_property (BaconVideoWidget *bvw,
                                       BaconVideoWidgetVideoProperty type)
{
  int value = 65535 / 2;
  g_return_val_if_fail (bvw != NULL, value);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), value);
  
  if (GST_IS_COLOR_BALANCE (bvw->priv->balance))
    {
      const GList *channels_list = NULL;
      GstColorBalanceChannel *found_channel = NULL;
      
      channels_list = gst_color_balance_list_channels (bvw->priv->balance);
      
      while (channels_list)
        { /* We search for the right channel corresponding to type */
          GstColorBalanceChannel *channel = channels_list->data;
          if (type == BVW_VIDEO_BRIGHTNESS && channel &&
              g_strrstr (channel->label, "BRIGHTNESS"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_CONTRAST && channel &&
              g_strrstr (channel->label, "CONTRAST"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_SATURATION && channel &&
              g_strrstr (channel->label, "SATURATION"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_HUE && channel &&
              g_strrstr (channel->label, "HUE"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          channels_list = g_list_next (channels_list);
        }
        
      if (GST_IS_COLOR_BALANCE_CHANNEL (found_channel))
        {
          value = gst_color_balance_get_value (bvw->priv->balance,
                                               found_channel);
          value = ((double) value - found_channel->min_value) * 65535 /
                  ((double) found_channel->max_value - found_channel->min_value);
          g_object_unref (found_channel);
        }
    }
    
  return value;
}

void
bacon_video_widget_set_video_property (BaconVideoWidget *bvw,
                                       BaconVideoWidgetVideoProperty type,
                                       int value)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  
  if ( !(value < 65535 && value > 0) )
    return;
  
  if (GST_IS_COLOR_BALANCE (bvw->priv->balance))
    {
      const GList *channels_list = NULL;
      GstColorBalanceChannel *found_channel = NULL;
      
      channels_list = gst_color_balance_list_channels (bvw->priv->balance);
      
      while (channels_list)
        { /* We search for the right channel corresponding to type */
          GstColorBalanceChannel *channel = channels_list->data;
          if (type == BVW_VIDEO_BRIGHTNESS && channel &&
              g_strrstr (channel->label, "BRIGHTNESS"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_CONTRAST && channel &&
              g_strrstr (channel->label, "CONTRAST"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_SATURATION && channel &&
              g_strrstr (channel->label, "SATURATION"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          else if (type == BVW_VIDEO_HUE && channel &&
              g_strrstr (channel->label, "HUE"))
            {
              g_object_ref (channel);
              found_channel = channel;
            }
          channels_list = g_list_next (channels_list);
        }
        
      if (GST_IS_COLOR_BALANCE_CHANNEL (found_channel))
        {
          value = value * ((double) found_channel->max_value - found_channel->min_value) / 65535 + found_channel->min_value;
          gst_color_balance_set_value (bvw->priv->balance, found_channel,
                                       value);
          g_object_unref (found_channel);
        }
    }
}

float
bacon_video_widget_get_position (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, -1);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), -1);
  return bvw->priv->current_position;
}

gint64
bacon_video_widget_get_current_time (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, -1);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), -1);
  return bvw->priv->current_time;
}

gint64
bacon_video_widget_get_stream_length (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, -1);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), -1);
  return bvw->priv->stream_length;
}

gboolean
bacon_video_widget_is_playing (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  if (GST_STATE (GST_ELEMENT (bvw->priv->play)) == GST_STATE_PLAYING)
    return TRUE;

  return FALSE;
}

gboolean
bacon_video_widget_is_seekable (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  /* hmm... */
  if (bvw->priv->stream_length)
    return TRUE;
  else
    return FALSE;
}

gboolean
bacon_video_widget_can_play (BaconVideoWidget * bvw, MediaType type)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);

  return FALSE;
}

G_CONST_RETURN gchar **
bacon_video_widget_get_mrls (BaconVideoWidget * bvw, MediaType type)
{
  g_return_val_if_fail (bvw != NULL, NULL);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), NULL);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), NULL);

  return NULL;
}

static void
bacon_video_widget_get_metadata_string (BaconVideoWidget * bvw,
					BaconVideoWidgetMetadataType type,
					GValue * value)
{
  char *string = NULL;
  gboolean res = FALSE;

  g_value_init (value, G_TYPE_STRING);

  if (bvw->priv->play == NULL)
    {
      g_value_set_string (value, NULL);
      return;
    }

  switch (type)
    {
    case BVW_INFO_TITLE:
      res = gst_tag_list_get_string (bvw->priv->tagcache,
				     GST_TAG_TITLE, &string);
      break;
    case BVW_INFO_ARTIST:
      res = gst_tag_list_get_string (bvw->priv->tagcache,
				     GST_TAG_ARTIST, &string);
      break;
    case BVW_INFO_YEAR:
      res = gst_tag_list_get_string (bvw->priv->tagcache,
				     GST_TAG_DATE, &string);
      break;
    case BVW_INFO_VIDEO_CODEC:
      res = gst_tag_list_get_string (bvw->priv->tagcache,
				     GST_TAG_VIDEO_CODEC, &string);
      break;
    case BVW_INFO_AUDIO_CODEC:
      res = gst_tag_list_get_string (bvw->priv->tagcache,
				     GST_TAG_AUDIO_CODEC, &string);
      break;
    case BVW_INFO_CDINDEX:
      res = FALSE;
      break;
    default:
      g_assert_not_reached ();
    }

  if (res)
    g_value_set_string (value, string);
  else
    g_value_set_string (value, NULL);

  return;
}

static void
bacon_video_widget_get_metadata_int (BaconVideoWidget * bvw,
				     BaconVideoWidgetMetadataType type,
				     GValue * value)
{
  int integer = 0;

  g_value_init (value, G_TYPE_INT);

  if (bvw->priv->play == NULL)
    {
      g_value_set_int (value, 0);
      return;
    }

  switch (type)
    {
    case BVW_INFO_DURATION:
      integer = bacon_video_widget_get_stream_length (bvw) / 1000;
      break;
    case BVW_INFO_DIMENSION_X:
      integer = bvw->priv->video_width;
      break;
    case BVW_INFO_DIMENSION_Y:
      integer = bvw->priv->video_height;
      break;
    case BVW_INFO_FPS:
      if (bvw->priv->video_fps - (int)bvw->priv->video_fps >= 0.5)
        integer = bvw->priv->video_fps + 1;
      else
        integer = bvw->priv->video_fps;
      break;
    case BVW_INFO_BITRATE:
      integer = 0;
      break;
    default:
      g_assert_not_reached ();
    }

  g_value_set_int (value, integer);

  return;
}

static void
bacon_video_widget_get_metadata_bool (BaconVideoWidget * bvw,
				      BaconVideoWidgetMetadataType type,
				      GValue * value)
{
  gboolean boolean = FALSE;

  g_value_init (value, G_TYPE_BOOLEAN);

  if (bvw->priv->play == NULL)
    {
      g_value_set_boolean (value, FALSE);
      return;
    }

  switch (type)
    {
    case BVW_INFO_HAS_VIDEO:
      boolean = bvw->priv->media_has_video;
      break;
    case BVW_INFO_HAS_AUDIO:
      boolean = bvw->priv->media_has_audio;
      break;
    default:
      g_assert_not_reached ();
    }

  g_value_set_boolean (value, boolean);

  return;
}

void
bacon_video_widget_get_metadata (BaconVideoWidget * bvw,
				 BaconVideoWidgetMetadataType type,
				 GValue * value)
{
  g_return_if_fail (bvw != NULL);
  g_return_if_fail (BACON_IS_VIDEO_WIDGET (bvw));
  g_return_if_fail (GST_IS_ELEMENT (bvw->priv->play));

  switch (type)
    {
    case BVW_INFO_TITLE:
    case BVW_INFO_ARTIST:
    case BVW_INFO_YEAR:
    case BVW_INFO_VIDEO_CODEC:
    case BVW_INFO_AUDIO_CODEC:
      bacon_video_widget_get_metadata_string (bvw, type, value);
      break;
    case BVW_INFO_DURATION:
    case BVW_INFO_DIMENSION_X:
    case BVW_INFO_DIMENSION_Y:
    case BVW_INFO_FPS:
    case BVW_INFO_BITRATE:
      bacon_video_widget_get_metadata_int (bvw, type, value);
      break;
    case BVW_INFO_HAS_VIDEO:
    case BVW_INFO_HAS_AUDIO:
      bacon_video_widget_get_metadata_bool (bvw, type, value);
      break;
    default:
      g_assert_not_reached ();
    }

  return;
}

/* Screenshot functions */
gboolean
bacon_video_widget_can_get_frames (BaconVideoWidget * bvw, GError ** error)
{
  g_return_val_if_fail (bvw != NULL, FALSE);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), FALSE);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), FALSE);
  return FALSE;
}

GdkPixbuf *
bacon_video_widget_get_current_frame (BaconVideoWidget * bvw)
{
  g_return_val_if_fail (bvw != NULL, NULL);
  g_return_val_if_fail (BACON_IS_VIDEO_WIDGET (bvw), NULL);
  g_return_val_if_fail (GST_IS_ELEMENT (bvw->priv->play), NULL);
  return NULL;
}

/* =========================================== */
/*                                             */
/*          Widget typing & Creation           */
/*                                             */
/* =========================================== */

G_DEFINE_TYPE(BaconVideoWidget, bacon_video_widget, GTK_TYPE_BOX)

struct poptOption *
bacon_video_widget_get_popt_table (void)
{
  /* Initializing GStreamer backend and parse our options from the command 
     line options */
  return (struct poptOption *) gst_init_get_popt_table ();
}

GQuark
bacon_video_widget_error_quark (void)
{
  static GQuark q = 0;
  if (q == 0) {
    q = g_quark_from_static_string ("bvw-error-quark");
  }
  return q;
}

GtkWidget *
bacon_video_widget_new (int width, int height,
			gboolean null_out, GError ** err)
{
  BaconVideoWidget *bvw;
  GstElement *audio_sink = NULL, *video_sink = NULL;

  bvw = BACON_VIDEO_WIDGET (g_object_new
                            (bacon_video_widget_get_type (), NULL));
  
  bvw->priv->play = gst_element_factory_make ("playbin", "play");
  if (!bvw->priv->play) {
    g_set_error (err, bacon_video_widget_error_quark (),
		 0, _("Failed to create a GStreamer play object"));
    g_object_unref (G_OBJECT (bvw));
    return NULL;
  }
  
  bvw->priv->media_device = g_strdup ("/dev/dvd");
  if (width <= 0 || height <= 0) {
    width = DEFAULT_WIDTH;
    height = DEFAULT_HEIGHT;
  }
  bvw->priv->init_width = width;
  bvw->priv->init_height = height;

  bvw->priv->cursor_shown = TRUE;
  bvw->priv->logo_mode = TRUE;
  bvw->priv->auto_resize = TRUE;

  audio_sink = gst_gconf_get_default_audio_sink ();
  if (!GST_IS_ELEMENT (audio_sink))
    {
      g_message ("failed to render default audio sink from gconf");
      return NULL;
    }
  video_sink = gst_gconf_get_default_video_sink ();
  if (!GST_IS_ELEMENT (video_sink))
    {
      g_message ("failed to render default video sink from gconf");
      return NULL;
    }

  g_object_set (G_OBJECT (bvw->priv->play), "video-sink",
		video_sink, NULL);
  g_object_set (G_OBJECT (bvw->priv->play), "audio-sink",
		audio_sink, NULL);

  bvw->priv->vis_plugins_list = NULL;

  g_signal_connect (G_OBJECT (bvw->priv->play), "eos",
		    (GtkSignalFunc) got_eos, (gpointer) bvw);
  g_signal_connect (G_OBJECT (bvw->priv->play), "state-change",
		    G_CALLBACK (state_change), (gpointer) bvw);
  g_signal_connect (G_OBJECT (bvw->priv->play), "found_tag",
		    (GtkSignalFunc) got_found_tag, (gpointer) bvw);
  g_signal_connect (G_OBJECT (bvw->priv->play), "error",
		    (GtkSignalFunc) got_error, (gpointer) bvw);

  /* We try to get an element supporting XOverlay interface */
  if (GST_IS_BIN (bvw->priv->play)) {
    GstElement *element = NULL;
    GList *elements = NULL, *l_elements = NULL;
    
    element = gst_bin_get_by_interface (GST_BIN (video_sink),
                                        GST_TYPE_X_OVERLAY);
    if (GST_IS_X_OVERLAY (element))
      bvw->priv->xoverlay = GST_X_OVERLAY (element);
    
    g_signal_connect (G_OBJECT (bvw->priv->xoverlay),
		      "desired-size-changed", G_CALLBACK (got_video_size),
		      (gpointer) bvw);
    
    /* Get them all and prefer hardware one */
    elements = gst_bin_get_all_by_interface (GST_BIN (bvw->priv->play),
                                             GST_TYPE_COLOR_BALANCE);
    l_elements = elements;
    
    /* We take the first one */
    if (elements && GST_IS_COLOR_BALANCE (elements->data))
      element = GST_ELEMENT (elements->data);
    
    while (elements)
      {
        GstElement *local_element = NULL;
        GstColorBalanceClass *cb_class = NULL;
        
        if (elements->data && GST_IS_COLOR_BALANCE (elements->data))
          {
            local_element = GST_ELEMENT (elements->data);
            cb_class = GST_COLOR_BALANCE_GET_CLASS (elements->data);
        
            /* If one of them is hardware type we use that one then */
            if (GST_COLOR_BALANCE_TYPE (cb_class) == GST_COLOR_BALANCE_HARDWARE)
              {
                element = local_element;
              }
          }
          
        elements = g_list_next (elements);
      }
    
    if (l_elements)
      g_list_free (l_elements);
    
    elements = l_elements = NULL;
    
    if (GST_IS_COLOR_BALANCE (element))
      {
        bvw->priv->balance = GST_COLOR_BALANCE (element);
      }
  }
  
  return GTK_WIDGET (bvw);
}
